package ru.ifmo.ga.ca;

import java.util.ArrayList;
import java.util.Arrays;
import java.io.Serializable;

/**
 * User: Yuri D. Bedny
 * Date: Apr 2, 2006
 * Time: 11:29:40 AM
 */

public class Rule implements Serializable {
    private int radius;
    private boolean[] f;

    private String info = "";

    public int getLength() {
        return 1 << (2 * radius + 1);
    }

    public int getRadius() {
        return radius;
    }

    public int cardinality() {
        int x = 0;
        for (boolean b : f)
            if (b)
                x++;
        return x;
    }

    public void setRadius(int radius) {
        this.radius = radius;
        f = new boolean[getLength()];
    }

    public boolean getBit(int i) {
        return f[i];
    }

    public void flipBit(int i) {
        f[i] = !f[i];
    }

    public boolean[] getTable() {
        return f;
    }

    public void setTable(boolean arr[]) {
        System.arraycopy(arr, 0, f, 0, arr.length);
    }

    public void setBit(int i, boolean b) {
        f[i] = b;
    }

    public void parseHex(String rule) throws RuntimeException {
        if (rule.length() * 4 != getLength()) {
            throw new RuntimeException("Expected string length " + getLength() + ", found " + rule.length());
        }
        for (int i = 0; i < rule.length(); i++) {
            int x = Integer.parseInt(String.valueOf(rule.charAt(i)), 16);
            for (int j = 0; j < 4; j++) {
                setBit(i * 4 + (3 - j), (x & 1) == 1);
                x >>= 1;
            }
        }
    }

    public void parseRule(String rule) throws RuntimeException {
        if (rule.length() != getLength()) {
            throw new RuntimeException("Expected string length " + getLength() + ", found " + rule.length());
        }
        for (int i = 0; i < getLength(); i++) {
            char bit = rule.charAt(i);
            if (bit != '0' && bit != '1') {
                throw new RuntimeException("Invalid character found at position " + i + ". String must contain only '0's and '1's.");
            }
            setBit(i, bit == '1');
        }
    }

    public String toString() {
        if (info.length() != 0)
            return info;

        StringBuilder b = new StringBuilder();
        for (int i = 0; i < getLength(); i++) {
            b.append(f[i] ? '1' : '0');
        }
        return b.toString();
    }


    public Rule(final Rule rule) {
        this.radius = rule.radius;
        this.f = new boolean[rule.f.length];
        System.arraycopy(rule.f, 0, this.f, 0, rule.f.length);
    }

    public Rule() {
    }

    public String getInfo() {
        return info;
    }

    public void setInfo(String info) {
        this.info = info;
    }

    public String toTable() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < getLength(); i++) {
            sb.append(Utils.toBinaryString(2 * radius + 1, i));
            sb.append(getBit(i) ? " 1\n" : " 0\n");
        }
        return sb.toString();
    }

    public int getMonotony() {
        int c = 0;
        for (int x1 = 0; x1 < getLength() - 1; x1++) {
            for (int x2 = x1 + 1; x2 < getLength(); x2++) {
                int y1 = f[x1] ? 1 : 0;
                int y2 = f[x2] ? 1 : 0;
                if (includesZero(x1, x2) && y1 > y2 || includesZero(x2, x1) && y2 > y1) {
                    c++;
                }
            }
        }
        return c;
    }

    public boolean isMonotonous() {
        return getMonotony() == 0;
    }

    public void randomize() {
        for (int i = 0; i < getLength(); i++) {
            f[i] = Math.random() > 0.5;
        }
    }

    /**
     * Automaton (rule) is washing iff linear shells of zero subsets intersection is an empty set.
     * @return if automaton (rule) is washing or not.
     */
    public boolean isWashing() {
        int l = 0;
        int h = 2 * radius;
        for (Integer x : getZeroSubsets()) {
            final int hz = highestZeroPos(x);    // leftmost
            final int lz = lowestZeroPos(x);     // rightmost
            l = Math.max(l, lz);
            h = Math.min(h, hz);
            if (l > h) {
                return true;
            }
        }
        return false;
    }

    private int highestZeroPos(int x) {//todo: very simple impl
        long k = 1 << (2 * radius);
        int pos = 0;
        while (k > 0 && (x & k) > 0) {
            pos++;
            k >>= 1;
        }
        return 2 * radius - pos;
    }

    private int lowestZeroPos(int x) { //todo: very simple impl
        int pos = 0;
        while ((x & 1) == 1) {
            pos++;
            x >>= 1;
        }
        return pos;
    }

    public ArrayList<Integer> getZeroSubsets() { //todo: stupid impl, optimize
        ArrayList<Integer> zeroSets = new ArrayList<Integer>();
        final boolean monotonous = isMonotonous();
        for (int x = 0; x < getLength(); x++) {
            if (f[x]) continue;
            if (monotonous || zeroOk(x)) {
                zeroSets.add(x);
            }
        }

        ArrayList<Integer> exc = new ArrayList<Integer>();
        for (int i = 0; i < zeroSets.size(); i++) {
            int x = zeroSets.get(i);
            for (int j = 0; j < zeroSets.size(); j++) {
                if (i == j) continue;
                int y = zeroSets.get(j);
                if (includesZero(y, x)) {
                    exc.add(y);
                }
            }
        }
        zeroSets.removeAll(exc);

        return zeroSets;
    }

    private boolean zeroOk(int x) {
        for (int y = 0; y < getLength(); y++) {
            if (includesZero(y, x) && f[y]) {
                return false;
            }
        }
        return true;
    }

    private boolean includesZero(int y, int x) {//todo: can be removed
        for (int i = 0; i < 2 * radius + 1; i++) {
            int dy = y & 1;
            int dx = x & 1;
            if (dx == 0 && dy == 1) {
                return false;
            }
            x >>= 1;
            y >>= 1;
        }
        return true;
    }

    private boolean includesOne(int y, int x) {//todo: optimize alg
        for (int i = 0; i < 2 * radius + 1; i++) {
            int dy = y & 1;
            int dx = x & 1;
            if (dx == 1 && dy == 0) {
                return false;
            }
            x >>= 1;
            y >>= 1;
        }
        return true;
    }

    public void invert() {
        for (int i = 0; i < getLength(); i++) {
            f[i] = !f[i];
        }
    }

    public String toGreyTable() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < getLength(); i++) {
            sb.append(f[((int) Utils.grey(i))] ? '1' : '0');
        }
        return sb.toString();
    }

    public String toCNF(char[] arr, String mul, String sum, String not, boolean isPrefixNot) {
        if (arr.length != 2 * radius + 1) {
            throw new IllegalArgumentException("arr.length != 2 * radius + 1");
        }

        StringBuilder sb =  new StringBuilder();
        boolean isEmpty = true;
        for (int i = 0; i < getLength(); i++) {
            if (!f[i]) continue;
            boolean b[] = Utils.getArr(2 * radius + 1, i);
            StringBuilder token = new StringBuilder();
            for (int j = 0; j < b.length; j++) {
                if (b[j]) {
                    token.append(arr[j]);
                } else {
                    token.append(isPrefixNot ? not + arr[j] : arr[j] + not);
                }
                if (j < b.length - 1) {
                    token.append(mul);
                }
            }
            if (!isEmpty) {
                sb.append(sum);
            }
            sb.append(token.toString());
            isEmpty = false;
        }

        return sb.toString();

    }

    public String toCNF(char[] arr) {
        return toCNF(arr, "", " + ", "'", false);
    }

    public String toWolframMathCNF(char[] arr) { // Wolfram's "Mathematica" notation
        return toCNF(arr, " && ", " || ", "!", true);
    }

    public int hashCode() {
        return Arrays.hashCode(f);
    }

    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }

        if (!(obj instanceof Rule)) {
            return false;
        }

        Rule r = (Rule) obj;
        return Arrays.equals(r.f, f);
    }

    public int getSymmetry() {
        int sym = 0;
        for (int i = 0; i < getLength(); i++) {

        }
        return sym;
    }
}

