package ru.ifmo.ga.ca;

import ru.ifmo.ga.ca.configuration.Configuration;
import ru.ifmo.ga.ca.configuration.ConfigurationFactory;

import java.util.ArrayList;

/**
 * User: Yuri D. Bedny
 * Date: Apr 2, 2006
 * Time: 10:56:05 AM
 */

/**
 * This class represents simple (one-dimensional, toroidal, binary-state) cellular automaton
 */
public class SimpleCA {
    private final ArrayList<CAListener> listeners = new ArrayList<CAListener>();
    private ConfigurationFactory.Implementation cfgImpl = ConfigurationFactory.Implementation.BooleanArray;
    private int size;
    private Configuration cfg;
    private Configuration f;

    private Rule rule;

    // cached
    private int lastNum;
    private boolean stopped;
    private int radius;
    private int oneMask;

    public SimpleCA() {

    }

    public ConfigurationFactory.Implementation getConfigurationImpl() {
        return cfgImpl;
    }

    private void updateF() {
        if (size > 0)
            f = new ConfigurationFactory(cfgImpl).create(size);
    }

    public void setConfigurationImpl(ConfigurationFactory.Implementation impl) {
        cfgImpl = impl;
        updateF();
    }

    public boolean isStopped() {
        return stopped;
    }

    public void setSize(int size) {
        this.size = size;
        updateF();
    }

    public int getSize() {
        return size;
    }

    public Rule getRule() {
        return rule;
    }

    public void setRule(Rule rule) {
        this.rule = rule;
        radius = rule.getRadius();
        oneMask = 1 << (2 * radius);
    }

    public void setConfiguration(Configuration c) {
        stopped = false;
        if (c.size() != size) {
            throw new RuntimeException("Expected configuration size " + size + ", found " + c.size());
        }
        cfg = (Configuration) c.clone();
    }

    public Configuration getConfiguration() {
        return cfg;
    }

    public boolean next() {
        if (stopped) {
            return false;
        }

        boolean eq = true;
        for (int i = 0; i < size; i++) {
            f.set(i, rule.getBit(getNum(i)));
            if (f.get(i) != cfg.get(i)) {
                eq = false;
            }
        }

        Configuration temp = cfg;
        cfg = f;
        f = temp;

        stopped = eq;
        if (stopped) {
            fireStopEvent(CAEvent.STOP);
        } else {
            fireNextEvent(CAEvent.NEXT);
        }
        return !eq;
    }

    private int getNum(int i) {
        int num = 0;
        if (i == 0) {
            for (int j = i + radius; j >= i - radius; j--) {
                int p = (j + size) % size;
                num = (num << 1) + (cfg.get(p) ? 1 : 0);
            }
        } else {
            int p = (i + radius) % size;
            num = (lastNum >> 1) + (cfg.get(p) ? oneMask : 0);
        }
        lastNum = num;
        return num;
    }

    public String toString() {
        return cfg.toString();
    }

    private void fireStopEvent(CAEvent e) {
        for (CAListener l : listeners) {
            l.stopped(e);
        }
    }

    private void fireNextEvent(CAEvent e) {
        for (CAListener l : listeners) {
            l.next(e);
        }
    }

    public void addCAListener(CAListener l) {
        listeners.add(l);
    }

    public void removeListener(CAListener l) {
        listeners.remove(l);
    }

}
