package ru.ifmo.ga.ca.dct;

import ru.ifmo.ga.ca.dct.modules.PointsSet;
import ru.ifmo.ga.fx.AbstractRepository;
import ru.ifmo.ga.fx.Point;
import ru.ifmo.ga.fx.UploadResult;
import ru.ifmo.ga.fx.desire.DownloadDesire;
import ru.ifmo.ga.fx.stat.Divergence;

import java.io.*;
import java.rmi.RemoteException;
import java.util.Date;
import java.util.ArrayList;

public class CARepository extends AbstractRepository {
    public final static String DEFAULT_CA_REPOSITORY_NAME = "ca_repository";

    public CARepository(String name) {
        super(name);
    }

    private final static int MAX_SIZE = 500;


    private final PointsSet points = new PointsSet();

    private final CAFitness fitness = new CAFitness();
    private final CADivergence divergence = new CADivergence();
    private int maxSize = MAX_SIZE;

    public CAFitness getFitness() throws RemoteException {
        return fitness;
    }

    public boolean removePoint(Point point) throws RemoteException {
        return points.remove(point);
    }

    public boolean replacePoint(Point oldPoint, Point newPoint) throws RemoteException {
        int size = points.size();
        CAPoint caNew = (CAPoint)(newPoint);
        CAPoint caOld = (CAPoint) oldPoint;

        boolean containsOld = points.remove(oldPoint);
        if (!containsOld) {
            System.out.println("[replace] !FAILED!");
            assert (size == points.size());
            return false;
        }

        points.remove(newPoint);
        boolean res = points.add(newPoint);
        if (res) {
            System.out.println("[replace] from " + caOld.getMetadata().getTestCases() + "\t" + caOld.getMetadata().getValue());
            System.out.println("[replace] to " + caNew.getMetadata().getTestCases() + "\t" + caNew.getMetadata().getValue());
            System.out.println();
        } else {
            System.out.println("[replace] !FAILED!");
        }
        assert (size == points.size());
        return res;
    }

    public Point downloadPoint(DownloadDesire desire) throws RemoteException {
        CAPoint p = (CAPoint)desire.downloadPoint(this);
        if (p != null) {
            System.out.println("- [" + desire.getClientName() + "] " + p.getMetadata().getValue() + ", " + p.getRule().toString());
        }
        return p;
    }

    private void add(CAPoint caPoint) {
        if (points.add(caPoint)) {
            System.out.println("+ [" + caPoint.getMetadata().getClientName() + "] " + caPoint.getMetadata().getValue() + ", " + caPoint.getRule().toString());
        }
    }

    public synchronized UploadResult uploadPoint(Point point) throws RemoteException {
        if (!(point instanceof CAPoint)) {
            return UploadResult.INVALID_TYPE;
        }

        CAPoint caPoint = (CAPoint) point;

        //todo: improve contains check
        if (points.contains(caPoint)) {
            return UploadResult.ALREADY_EXISTS;
        }

        //todo: actually UploadPointDesire is needed to specify that we don't want calculating fintness on server side
        if (caPoint.getMetadata().getValue() == 0) {
            System.out.println("[uploadPoint] calculating...");
            double value = fitness.calculate(caPoint);
            caPoint.getMetadata().setValue(value);
            caPoint.getMetadata().setTestCases(fitness.getTestCases());
        }
        
        if (points.size() < maxSize) {
            add(caPoint);
            return UploadResult.OK;
        }

        if (caPoint.getMetadata().getValue() < points.first().getMetadata().getValue()) {
            return UploadResult.TOO_WEEK;
        }

        add(caPoint);
        return UploadResult.OK;
    }

    public int getMaxSize() throws RemoteException {
        return maxSize;
    }

    public void setMaxSize(int size) throws RemoteException {
        maxSize = size;
    }

    public PointsSet getPoints() throws RemoteException {
        return points;
    }

    public double calculate(Point p) throws RemoteException  {
        return fitness.calculate(p);
    }

    public void save() throws IOException {
        System.out.print("[" + new Date().toString() + "]" + " Saving repository '" + getName() +  "' ... ");
        DataOutputStream dos = new DataOutputStream(new FileOutputStream(getName()));
        dos.writeInt(maxSize);
        dos.writeInt(points.size());
        for (Point p : points) {
            ((CAPoint)p).save(dos);
        }
        dos.close();
        System.out.println("ok!");
    }

    public void load() throws IOException {
        System.out.println("Loading repository " + getName() + "... ");
        DataInputStream dis = new DataInputStream(new FileInputStream(getName()));
        maxSize = dis.readInt();
        int size = dis.readInt();
        points.clear();
        for (int i = 0; i < size; i++) {
            CAPoint caPoint = new CAPoint();
            caPoint.load(dis);
            if (points.add(caPoint)) {
                System.out.println(caPoint.getMetadata().getTestCases() + " " + caPoint.getMetadata().getValue());
            }
        }
        dis.close();

        System.out.println("ok! size: " + size);
        if (getSize() > 0) {
            System.out.println("max: " + getMaxFitness() + ", min: " + getMinFitness() + ", avg: " + getAvgFitness() + ", div: " + getAvgDivergence());
        }


    }

    public double getMaxFitness() throws RemoteException {
        return points.last().getMetadata().getValue();
    }

    public double getMinFitness() throws RemoteException {
        return  points.first().getMetadata().getValue();
    }

    public double getAvgDivergence() throws RemoteException {
        double sum = 0;
        final int n = points.size();
        CAPoint [] arr = new CAPoint[n];
        points.toArray(arr);
        for (int i = 0; i < n - 1; i++) {
            for (int j = i + 1; j < n; j++) {
                sum += divergence.getDivergence(arr[i], arr[j]);
            }
        }
        return 2 * sum / (n * (n-1));
    }

    public double getAvgFitness() throws RemoteException {
        double sum = 0;
        for (Point p : points) {
            sum += p.getMetadata().getValue();
        }
        return sum / points.size();
    }

    public int getSize() throws RemoteException {
        return points.size();
    }

    public Divergence getDivergence() {
        return divergence;
    }

    public void clear() {
        points.clear();
    }

    public ArrayList<Point> downloadPoints(DownloadDesire desire) throws RemoteException {
        ArrayList<Point> points = desire.downloadPoints(this);
        if (points != null) {
            System.out.println("- [" + desire.getClientName() + "] " + points.size());
        }
        return points;
    }
}

