package ru.ifmo.ga.ca.tree;

import ext.bmin.ProductTerm;

import java.util.ArrayList;

import ru.ifmo.ga.ca.Rule;

/**
 * User: Yuri D. Bedny
 * Date: May 29, 2006
 * Time: 1:40:31 AM
 */

public class Tree {     //mutable
    private int varNum;

    private Tree tThen;
    private Tree tElse;

    private boolean isTruth;
    private boolean isFalse;
    private char[] vars;

    public boolean isTerminal() {
        return isTruth || isFalse;
    }

    public Tree(boolean isTruth, char[] vars) {
        this.vars = vars;
        this.isTruth = isTruth;
        this.isFalse = !isTruth;
    }

    public Tree(Rule r, char[] vars) {
        this(new Expression(r, vars));
    }

    public Tree(String s) {
        this(new Expression(s));
    }

    public Tree(Expression expr) {
        setExpression(expr);
    }

    private void build(Expression expr) {
        ArrayList<ProductTerm> thenTerms = new ArrayList<ProductTerm>();
        ArrayList<ProductTerm> elseTerms = new ArrayList<ProductTerm>();
        varNum = defineVar(expr);
        for (ProductTerm term : expr.getTerms()) {
            if (term.hasVar(varNum)) {
                ProductTerm newTerm = new ProductTerm(term);
                newTerm.excludeVar(varNum);
                if (term.hasNot(varNum)) {
                    elseTerms.add(newTerm);
                } else {
                    thenTerms.add(newTerm);
                }
            } else {
                thenTerms.add(new ProductTerm(term));
                elseTerms.add(new ProductTerm(term));
            }
        }
        tThen = new Tree(new Expression(thenTerms, expr.getVars()));
        tElse = new Tree(new Expression(elseTerms, expr.getVars()));
    }

    private int defineVar(Expression expr) {//todo: optimize if needed
        int max = 0;
        int maxi = -1;
        for (int i = 0; i < expr.getVarsNum(); i++) {
            final int k = count(expr, i);
            if (k > max) {
                max = k;
                maxi = i;
            }
        }
        return maxi;
    }

    private int count(Expression expr, int i) {
        int k = 0;
        for (ProductTerm term : expr.getTerms()) {
            if (term.hasVar(i)) {
                k++;
            }
        }
        return k;
    }

    public boolean calc(boolean []values) {
        if (isTruth) {
            return true;
        } else if (isFalse) {
            return false;
        }

        return values[values.length - varNum - 1] ? tThen.calc(values) : tElse.calc(values);
    }

    public Tree getThen() {
        return tThen;
    }

    public Tree getElse() {
        return tElse;
    }

    public int getVarNum() {
        return varNum;
    }

    public char getVarChar() {
        return vars[vars.length - varNum - 1];
    }

    public String toString() {
        if (isTerminal()) {
            return isTruth ? "1" : "0";
        }
        StringBuilder sb = new StringBuilder();
        sb.append(getVarChar());
        sb.append("(");
        sb.append(tThen.toString());
        sb.append(", ");
        sb.append(tElse.toString());
        sb.append(")");
        return sb.toString();
    }

    public int size() {
        if (isTerminal()) {
            return 1;
        }

        return 1 + tThen.size() + tElse.size();
    }

    public int avgHeight() {
        if (isTerminal()) {
            return 1;
        }

        return (tThen.height() + tElse.height()) / 2;
    }

    public int height() {
        if (isTerminal()) {
            return 1;
        }

        return 1 + Math.max(tThen.height(), tElse.height());
    }

    public Rule toRule() {
        Rule r = new Rule();
        final int d = vars.length;
        r.setRadius(d / 2);
        final int max = 1 << d;
        for (int i = 0; i < max; i++) {
            r.setBit(i, calc(getArr(d, i)));
        }
        return r;
    }

    private boolean[] getArr(int size, int x) {
        boolean[] a = new boolean[size];
        for (int i = 0; i < size; i++) {
            a[i] = (x & 1) == 1;
            x >>= 1;
        }
        return a;
    }

    public void setExpression(Expression expr) {
        this.isTruth = expr.isTruth();
        this.isFalse = expr.isFalse();
        this.vars = expr.getVars();
        if (!isTruth && !isFalse) {
            build(expr);
        } else {
            this.tThen = null;
            this.tElse = null;
        }
    }

    public void setVarNum(int varNum) {
        this.varNum = varNum;
    }

    public void setThen(Tree tThen) {
        this.tThen = tThen;
    }

    public void setElse(Tree tElse) {
        this.tElse = tElse;
    }

    public boolean isTruth() {
        return isTruth;
    }

    public boolean isFalse() {
        return isFalse;
    }

    public char[] getVars() {
        return vars;
    }

    public void setTerminal(boolean isTruth, boolean isFalse) {
        this.isTruth = isTruth;
        this.isFalse = isFalse;
    }

    public ArrayList<Tree> getNodes() {
        ArrayList<Tree> nodes = new ArrayList<Tree>();
        nodes.add(this);
        if (!isTerminal()) {
            nodes.addAll(tThen.getNodes());
            nodes.addAll(tElse.getNodes());
        }
        return nodes;
    }

    public void tryFlipTerminal(boolean isTruth) {
        if (!isTerminal()) {
            this.isTruth = isTruth;
            this.isFalse = !isTruth;
        } else {
            if (tThen != null && tElse != null) {
                this.isTruth = false;
                this.isFalse = false;
            }
        }
    }

}
