package ru.ifmo.ga.ca.dct;

import junit.framework.TestCase;
import ru.ifmo.ga.ca.KnownCARepository;
import ru.ifmo.ga.ca.Utils;
import ru.ifmo.ga.ca.Rule;
import ru.ifmo.ga.fx.Point;
import ru.ifmo.ga.fx.Repository;
import ru.ifmo.ga.fx.UploadResult;
import ru.ifmo.ga.fx.desire.SlightlyBetterPointDesire;

public class CARepositoryTest extends TestCase {
    private Repository rep;
    private DCTServer server;
    private final static String REP_NAME = "test_ca_repository";

    public void setUp() throws Exception {
        server = new DCTServer();
        server.start();
        Repository testRepository = new CARepository(REP_NAME);
        try {
            server.createRepository(testRepository);
        } catch (Exception e) {

        }

        rep = server.getRepository(REP_NAME);
        rep.clear();
        assertEquals(0, rep.getSize());
    }


    public void testUploadEqualsRules() throws Exception {
        Rule r1 = Utils.randomRule();
        Rule r2 = new Rule(r1);

        CAPoint ca1 = new CAPoint();
        ca1.setRule(r1);
        assertEquals(UploadResult.OK, rep.uploadPoint(ca1));

        CAPoint ca2 = new CAPoint();
        ca2.setRule(r2);
        assertEquals(UploadResult.ALREADY_EXISTS, rep.uploadPoint(ca2));

        CAPoint ca3 = new CAPoint();
        ca3.setRule(r2);
        assertEquals(UploadResult.ALREADY_EXISTS, rep.uploadPoint(ca3));

        r2.flipBit(1);
        r2.flipBit(100);
        assertEquals(UploadResult.OK, rep.uploadPoint(ca3));
    }

    public void testCreateWithSameName() throws Exception {
        Repository testRepository = new CARepository(REP_NAME);
        try {
            server.createRepository(testRepository);
            assertFalse(true);
        } catch (Exception e) {
        }
    }

    public void testCARepository() throws Exception{
        assertNotNull(rep);
        assertEquals(REP_NAME, rep.getName());
    }

    public void testCARepositoryFitness() throws Exception {
        CAPoint caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.JP_2));
        assertEquals(0.86, rep.calculate(caPoint), 0.2);
    }

    public void testUploadWeekCA() throws Exception {
        final int MS = 2;
        rep.setMaxSize(MS);
        assertEquals(MS, rep.getMaxSize());

        CAPoint caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.JP_1));
        assertEquals(UploadResult.OK, rep.uploadPoint(caPoint));

        caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.JP_2));
        assertEquals(UploadResult.OK, rep.uploadPoint(caPoint));

        caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.GKL));
        assertEquals(UploadResult.TOO_WEEK, rep.uploadPoint(caPoint));

    }

    public void testUploadGoodCA() throws Exception {
        final int MS = 2;
        rep.setMaxSize(MS);
        assertEquals(MS, rep.getMaxSize());

        CAPoint caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.GKL));
        assertEquals(UploadResult.OK, rep.uploadPoint(caPoint));

        caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.GEP_1));
        assertEquals(UploadResult.OK, rep.uploadPoint(caPoint));

        caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(KnownCARepository.JP_1));
        assertEquals(UploadResult.OK, rep.uploadPoint(caPoint));

    }

    private CAPoint getCA(String rule) throws Exception {
        CAPoint caPoint = new CAPoint();
        caPoint.setRule(Utils.getRule(rule));
        return caPoint;
    }

    public void testDownloadSlightlyBetter() throws Exception {
        rep.uploadPoint(getCA(KnownCARepository.GKL));
        CAPoint caPoint = (CAPoint) rep.downloadPoint(new SlightlyBetterPointDesire(getCA(KnownCARepository.GEP_2)));
        assertNull(caPoint);

        rep.uploadPoint(getCA(KnownCARepository.JP_2));
        caPoint = (CAPoint) rep.downloadPoint(new SlightlyBetterPointDesire(getCA(KnownCARepository.GEP_2)));
        assertEquals(caPoint.getRule(), getCA(KnownCARepository.JP_2).getRule());

        caPoint = (CAPoint) rep.downloadPoint(new SlightlyBetterPointDesire(getCA(KnownCARepository.MY_3)));
        assertEquals(caPoint.getRule(), getCA(KnownCARepository.GKL).getRule());
    }

    public void testSaveLoad() throws Exception {
        // save
        rep.setMaxSize(2);

        CAPoint caPoint = getCA(KnownCARepository.JP_1);
        caPoint.getMetadata().setValue(3);
        rep.uploadPoint(caPoint);

        caPoint = getCA(KnownCARepository.JP_2);
        caPoint.getMetadata().setValue(1);
        rep.uploadPoint(caPoint);

        rep.save();

        // change
        rep.getPoints().clear();

        // load
        rep.load();
        assertEquals(2, rep.getMaxSize());
        assertEquals(2, rep.getPoints().size());
        int i = 0;
        for (Point p : rep.getPoints()) {
            caPoint = (CAPoint) p;
            if (i == 1) {
                assertEquals(caPoint.getRule().toString(), KnownCARepository.JP_1);
                assertEquals(3, caPoint.getMetadata().getValue(), 0.01);
            } else {
                assertEquals(caPoint.getRule().toString(), KnownCARepository.JP_2);
                assertEquals(1, caPoint.getMetadata().getValue(), 0.01);
            }

            i++;
        }

    }

}
